////////////////////////////////////////////////////////////////////////////////
//                  ObjectLab is sponsoring QALab
// 
// Based in London, we are world leaders in the design and development 
// of bespoke applications for the Securities Financing markets.
// 
// <a href="http://www.objectlab.co.uk/open">Click here to learn more</a>
//           ___  _     _           _   _          _
//          / _ \| |__ (_) ___  ___| |_| |    __ _| |__
//         | | | | '_ \| |/ _ \/ __| __| |   / _` | '_ \
//         | |_| | |_) | |  __/ (__| |_| |__| (_| | |_) |
//          \___/|_.__// |\___|\___|\__|_____\__,_|_.__/
//                   |__/
//
//                   http://www.ObjectLab.co.uk
// ---------------------------------------------------------------------------
//
//QALab is released under the GNU General Public License.
//
//QALab: Collects QA Statistics from your build over time.
//2005+, ObjectLab Ltd
//
//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.
//
//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//General Public License for more details.
//
//You should have received a copy of the GNU General Public
//License along with this library; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
////////////////////////////////////////////////////////////////////////////////
package net.objectlab.qalab.ant;

import net.objectlab.qalab.interfaces.QALabExporter;
import net.objectlab.qalab.parser.StatMerger;
import net.objectlab.qalab.util.TaskLogger;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;
import org.xml.sax.InputSource;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

/**
 * The ant task that handles the merger of statistics into qalab.xml.
 * 
 * @author Benoit Xhenseval
 * @version $Revision: 187 $
 */
public class BuildStatMergeTask extends Task {
    // ~ Instance fields
    // ------------------------------------------------------------------------

    /** the base properties file. */
    private File inputFile = null;

    /** the merged properties file. */
    private File outputFile = null;

    /** if true, logging output will be suppressed. */
    private boolean quiet = false;

    /** the directory where the source code is. */
    private String srcDir;

    /** the handler for these statistics. */
    private String handler;

    /** a given timestamp for the stats. */
    private String mergerTimeStamp;

    /** exporter class name. */
    private String exporterClassName = "net.objectlab.qalab.exporter.QALabXMLExporter";

    /** properties file to use instead of setting them one by one. */
    private File propertiesFile;

    /** the loaded properties. */
    private Properties properties = null;

    // ~ Methods
    // ------------------------------------------------------------------------

    /**
     * Sets the class name for the handler for the given statistics.
     * 
     * @param theHandler
     *            class name for the handler.
     */
    public final void setHandler(final String theHandler) {
        this.handler = theHandler;
    }

    /**
     * @return the handler
     */
    private String getHandler() {
        return handler;
    }

    /**
     * @param sourceDirectory
     *            directory where the source code is.
     */
    public final void setSrcDir(final String sourceDirectory) {
        this.srcDir = sourceDirectory;
    }

    /**
     * Sets the input file containing the new statistics (checkstyle,etc).
     * 
     * @param newStatisticsFile
     *            the whose values will be incorporated in the qalab.xml
     */
    public final void setInputFile(final File newStatisticsFile) {
        this.inputFile = newStatisticsFile;
    }

    /**
     * Set the to output file (typically qalab.xml).
     * 
     * @param consolidatedFile
     *            the destination file where the statistics are written. The
     *            file may or may not exists.
     */
    public final void setOutputFile(final File consolidatedFile) {
        this.outputFile = consolidatedFile;
    }

    /**
     * Sets the quiet mode.
     * 
     * @param noLog
     *            if true, no messages are displayed while merging stats.
     */
    public final void setQuiet(final boolean noLog) {
        this.quiet = noLog;
    }

    /**
     * Sets the timestamp mode.
     * 
     * @param timestamp
     *            if present, uses this as timestamp for stats.
     */
    public final void setMergerTimeStamp(final String timestamp) {
        this.mergerTimeStamp = timestamp;
    }

    /**
     * Method invoked by the ant framework to execute the action associated with
     * this task. This will validate the input parameters then merge the
     * statistics.
     */
    public final void execute() {
        properties = createOverridingProperties();

        // validate the provided parameters
        validate();

        // Display the files being processed
        if (!quiet) {
            log("inputFile='" + inputFile.getPath() + "', outputFile='"
                    + outputFile.getPath());
            log("srcDir='" + srcDir + "', mergerTimeStamp=" + mergerTimeStamp);
            log("quiet='"
                    + quiet
                    + "', propertiesFile="
                    + (propertiesFile != null ? propertiesFile.getPath() : null));
            for (Iterator it = properties.entrySet().iterator(); it.hasNext();) {
                final Map.Entry entry = (Map.Entry) it.next();
                final String key = entry.getKey().toString();
                if (key.indexOf("qalab") >= 0) {
                    log(key + " = '" + entry.getValue() + "'");
                }
            }
        }

        mergeFiles();
    }

    /**
     * Merge the statistics into the qalab.xml.
     */
    private void mergeFiles() {
        try {
            TaskLogger logger = new AntTaskLogger(this);

            // create the exporter
            QALabExporter exporter = (QALabExporter) Class.forName(
                    getExporterClassName()).newInstance();

            properties.setProperty("qalab.merge.output.file", outputFile
                    .getAbsolutePath());

            exporter.setQuiet(quiet);
            exporter.setTaskLogger(logger);

            StatMerger merger = (StatMerger) Class.forName(getHandler())
                    .newInstance();

            merger.setQuiet(quiet);
            merger.setSrcDir(srcDir);
            merger.setTaskLogger(logger);
            final String tsOnly = properties
                    .getProperty("qalab.merge.timestampdateonly");
            merger.setMergerTimeStamp(mergerTimeStamp, tsOnly != null
                    && (tsOnly.equalsIgnoreCase("yes") || tsOnly
                            .equalsIgnoreCase("true")));

            properties.setProperty("qalab.merge.output.timestamp", merger
                    .getMergerTimeStamp());
            properties.setProperty("qalab.merge.type", merger.getType());
            exporter.configure(properties);

            merger.mergeStats(new InputSource(new FileReader(inputFile)),
                    exporter);
            log("Files: " + merger.getFileCount() + " Statistics:"
                    + merger.getTotalStatistics());

            exporter.save();
        } catch (IllegalAccessException e) {
            e.printStackTrace();
            throw new BuildException(e.toString());
        } catch (FileNotFoundException e) {
            e.printStackTrace();
            throw new BuildException(e.toString());
        } catch (IOException e) {
            e.printStackTrace();
            throw new BuildException(e.toString());
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
            throw new BuildException(e.toString());
        } catch (InstantiationException e) {
            e.printStackTrace();
            throw new BuildException(e.toString());
        } catch (Exception e) {
            e.printStackTrace();
            throw new BuildException(e.toString());
        }
    }

    /**
     * Validates the parameters supplied by ant.
     */
    private void validate() {
        if (inputFile == null) {
            throw new BuildException("inputFile is mandatory");
        }

        if (!inputFile.canRead()) {
            final String message = "Unable to read from " + inputFile + ".";

            throw new BuildException(message);
        }
    }

    /**
     * @param exporterClass
     *            The exporterClassName to set.
     */
    public final void setExporterClassName(final String exporterClass) {
        this.exporterClassName = exporterClass;
    }

    /**
     * @return Returns the exporterClassName.
     */
    public final String getExporterClassName() {
        if (exporterClassName != null) {
            return exporterClassName;
        } else {
            return properties.getProperty("qalab.merge.exporterclass");
        }
    }

    /**
     * Sets a properties file for use instead of individually setting them.
     * 
     * @param aProps
     *            the properties File to use
     */
    public final void setPropertiesFile(final File aProps) {
        propertiesFile = aProps;
    }

    /**
     * Create the Properties object based on the arguments specified to the ANT
     * task.
     * 
     * @return the properties for property expansion expansion
     */
    private Properties createOverridingProperties() {
        final Properties retVal = new Properties();

        // Load the properties file if specified
        if (propertiesFile != null && propertiesFile.canRead()
                && propertiesFile.isFile()) {
            FileInputStream inStream = null;
            if (!quiet) {
                log("loading " + propertiesFile);
            }
            try {
                inStream = new FileInputStream(propertiesFile);
                retVal.load(inStream);
            } catch (FileNotFoundException e) {
                throw new BuildException("Could not find Properties file '"
                        + propertiesFile + "'", e, getLocation());
            } catch (IOException e) {
                throw new BuildException("Error loading Properties file '"
                        + propertiesFile + "'", e, getLocation());
            } finally {
                try {
                    if (inStream != null) {
                        inStream.close();
                    }
                } catch (IOException e) {
                    throw new BuildException("Error closing Properties file '"
                            + propertiesFile + "'", e, getLocation());
                }
            }
        }

        // override with Ant properties like ${basedir}
        final Hashtable antProps = this.getProject().getProperties();

        for (Iterator it = antProps.entrySet().iterator(); it.hasNext();) {
            final Map.Entry entry = (Map.Entry) it.next();
            final String value = String.valueOf(entry.getValue());
            if (!quiet && entry.getKey().toString().indexOf("qalab") >= 0) {
                log("Adding Prop: " + entry.getKey() + " / [" + value + "]");
            }
            retVal.put(entry.getKey(), value);
        }

        return retVal;
    }
}
/*
 *                   ObjectLab is sponsoring QALab
 * 
 * Based in London, we are world leaders in the design and development 
 * of bespoke applications for the securities financing markets.
 * 
 * <a href="http://www.objectlab.co.uk/open">Click here to learn more about us</a>
 *           ___  _     _           _   _          _
 *          / _ \| |__ (_) ___  ___| |_| |    __ _| |__
 *         | | | | '_ \| |/ _ \/ __| __| |   / _` | '_ \
 *         | |_| | |_) | |  __/ (__| |_| |__| (_| | |_) |
 *          \___/|_.__// |\___|\___|\__|_____\__,_|_.__/
 *                   |__/
 *
 *                     www.ObjectLab.co.uk
 */
