////////////////////////////////////////////////////////////////////////////////
//
//                  ObjectLab is sponsoring QALab
// 
// Based in London, we are world leaders in the design and development 
// of bespoke applications for the Securities Financing markets.
// 
// <a href="http://www.objectlab.co.uk/open">Click here to learn more</a>
//           ___  _     _           _   _          _
//          / _ \| |__ (_) ___  ___| |_| |    __ _| |__
//         | | | | '_ \| |/ _ \/ __| __| |   / _` | '_ \
//         | |_| | |_) | |  __/ (__| |_| |__| (_| | |_) |
//          \___/|_.__// |\___|\___|\__|_____\__,_|_.__/
//                   |__/
//
//                   http://www.ObjectLab.co.uk
// ---------------------------------------------------------------------------
//
//QALab is released under the GNU General Public License.
//
//QALab: Collects QA Statistics from your build over time.
//2005+, ObjectLab Ltd
//
//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.
//
//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//General Public License for more details.
//
//You should have received a copy of the GNU General Public
//License along with this library; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
////////////////////////////////////////////////////////////////////////////////
package net.objectlab.qalab.ant;

import net.objectlab.qalab.parser.BuildStatMoverHandler;
import net.objectlab.qalab.util.QALabTags;
import net.objectlab.qalab.util.TaskLogger;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.parsers.ParserConfigurationException;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;

import java.text.ParseException;

/**
 * An Ant task to create the mover XML file given some parameters.
 * 
 * @author Paramjit Rehinsi
 * @version $Revision: 187 $
 */
public class BuildStatMoverTask extends Task implements TaskLogger {
    /** size of the string based on ISO 8601 date format. */
    private static final int MIN_DATE_SIZE = 18;

    /**
     * the inputFile, typically qalab.xml.
     */
    private File inputFile = null;

    /**
     * number of hours to define the time window from now.
     */
    private String startTimeHoursOffset;

    /**
     * Statistic types to be taken into account
     * (checkstyle|findbugs|pms|simian). Or comma separated combinations.
     */
    private String types;

    /**
     * The output XML file with ups and downs.
     */
    private File outputXMLfile;

    /**
     * if true, no output log.
     */
    private boolean quiet = false;

    /**
     * if true, adjust for weekend ie add 48h if start time falls on a weekend.
     */
    private boolean weekendAdjustment = false;

    /**
     * Start Time Window (instead of using startTimeHoursOffset).
     */
    private String startTimeWindow = null;

    /**
     * End Time Window (instead of assuming "now").
     */
    private String endTimeWindow = null;

    /**
     * offset in hours to determine if the stat has fallen to 0 over the last n
     * hours from the endDateCutoff.
     */
    private Integer hoursOffsetForLastRun = null;

    /**
     * @return true if no log enabled.
     */
    public final boolean isQuiet() {
        return quiet;
    }

    /**
     * @param noLog
     *            true and the system will not output debug info.
     */
    public final void setQuiet(final boolean noLog) {
        this.quiet = noLog;
    }

    /**
     * Method invoked by the ant framework to execute the compilation of
     * statistics movers withing the time window.
     */
    public final void execute() {
        validate();

        if (!quiet) {
            log("inputFile='" + inputFile.getPath() + "', type='" + getTypes() + "'" + " output='" + getOutputXMLfile().getPath()
                    + "' adjustForWeekend='" + isWeekendAdjustment() + "'" + " startTimeOffset='" + getStartTimeHoursOffset()
                    + "'" + " startTimeWindow='" + getStartTimeWindow() + "'" + " endTimeWindow='" + getEndTimeWindow() + "'");
        }

        try {
            final BuildStatMoverHandler handler = new BuildStatMoverHandler(new InputSource(new FileReader(getInputFile())),
                    getTypes(), new FileWriter(getOutputXMLfile()), isWeekendAdjustment(), this, quiet);

            if (startTimeWindow != null) {
                handler.setStartTimeWindow(QALabTags.DEFAULT_DATETIME_FORMAT.parse(startTimeWindow));
            }

            if (endTimeWindow != null) {
                handler.setEndTimeWindow(QALabTags.DEFAULT_DATETIME_FORMAT.parse(endTimeWindow));
            }

            if (getStartTimeHoursOffset() != null) {
                handler.setOffsetHours(getStartTimeHoursOffset());
            }

            if (getHoursOffsetForLastRun() != null) {
                handler.setHoursOffsetForLastRun(hoursOffsetForLastRun.intValue());
            }

            handler.process();
        } catch (ParserConfigurationException e) {
            log(e.toString());
        } catch (SAXException e) {
            log(e.toString());
        } catch (IOException e) {
            log(e.toString());
        } catch (ParseException e) {
            throw new BuildException("Date Parse issue", e);
        }
    }

    /**
     * Do some basic parameter validation.
     */
    private void validate() {
        if (inputFile == null) {
            throw new BuildException("inputFile is mandatory");
        }

        if ((startTimeHoursOffset != null) && (startTimeWindow != null)) {
            throw new BuildException("use startTimeHoursOffset OR " + "startTimeWindow, not both.");
        }
    }

    /**
     * uses the super log from the conventional Ant Task log.
     * 
     * @param text
     *            the text to log
     */
    public final void log(final String text) {
        super.log(text);
    }

    /**
     * @return Return the offset time in hours.
     */
    public final String getStartTimeHoursOffset() {
        return startTimeHoursOffset;
    }

    /**
     * Set the offset time in hours to define start window.
     * 
     * @param hours
     *            offset in hours
     */
    public final void setStartTimeHoursOffset(final String hours) {
        startTimeHoursOffset = hours;
    }

    /**
     * @return Return List of types handled (checkstyle|findbugs|pmd|simian).
     */
    public final String getTypes() {
        return types;
    }

    /**
     * Sets the types on which the report is based, checkstyle|findbugs|pmd
     * |simian can be combined in a comma separated format.
     * 
     * @param type
     *            single or multiple types.
     */
    public final void setTypes(final String type) {
        this.types = type;
    }

    /**
     * @return Return the destination XML file for the up/down stats.
     */
    public final File getOutputXMLfile() {
        return outputXMLfile;
    }

    /**
     * Set the destination XML file for the up/down stats.
     * 
     * @param outputFile
     *            XML destination file.
     */
    public final void setOutputXMLfile(final File outputFile) {
        outputXMLfile = outputFile;
    }

    /**
     * @return the input base file xml document, typically qalab.xml.
     */
    public final File getInputFile() {
        return inputFile;
    }

    /**
     * Set the base file xml document, typically qalab.xml.
     * 
     * @param inputXML
     *            the qalab.xml.
     */
    public final void setInputFile(final File inputXML) {
        inputFile = inputXML;
    }

    /**
     * @param wkendadj
     *            true will deduce a further 48 hours to the start window in
     *            order to take into account weekend if the start window fell on
     *            a Saturday or Sunday.
     */
    public final void setWeekendAdjustment(final boolean wkendadj) {
        weekendAdjustment = wkendadj;
    }

    /**
     * @return true if system is supposed to adjust for weekend.
     */
    public final boolean isWeekendAdjustment() {
        return weekendAdjustment;
    }

    /**
     * @param endTime
     *            The endTimeWindow to set.
     */
    public final void setEndTimeWindow(final String endTime) {
        this.endTimeWindow = endTime.trim();

        if (endTimeWindow.length() < MIN_DATE_SIZE) {
            endTimeWindow += " 00:00:00";
        }
    }

    /**
     * @return Returns the endTimeWindow.
     */
    public final String getEndTimeWindow() {
        return endTimeWindow;
    }

    /**
     * @param startTime
     *            The startTimeWindow to set.
     */
    public final void setStartTimeWindow(final String startTime) {
        this.startTimeWindow = startTime.trim();

        if (startTimeWindow.length() < MIN_DATE_SIZE) {
            startTimeWindow += " 00:00:00";
        }
    }

    /**
     * @return Returns the startTimeWindow.
     */
    public final String getStartTimeWindow() {
        return startTimeWindow;
    }

    public Integer getHoursOffsetForLastRun() {
        return hoursOffsetForLastRun;
    }

    public void setHoursOffsetForLastRun(Integer hoursOffsetForLastRun) {
        this.hoursOffsetForLastRun = hoursOffsetForLastRun;
    }
}
/*
 *                   ObjectLab is sponsoring QALab
 * 
 * Based in London, we are world leaders in the design and development 
 * of bespoke applications for the securities financing markets.
 * 
 * <a href="http://www.objectlab.co.uk/open">Click here to learn more about us</a>
 *           ___  _     _           _   _          _
 *          / _ \| |__ (_) ___  ___| |_| |    __ _| |__
 *         | | | | '_ \| |/ _ \/ __| __| |   / _` | '_ \
 *         | |_| | |_) | |  __/ (__| |_| |__| (_| | |_) |
 *          \___/|_.__// |\___|\___|\__|_____\__,_|_.__/
 *                   |__/
 *
 *                     www.ObjectLab.co.uk
 */
