////////////////////////////////////////////////////////////////////////////////
//
//                  ObjectLab is sponsoring QALab
// 
// Based in London, we are world leaders in the design and development 
// of bespoke applications for the Securities Financing markets.
// 
// <a href="http://www.objectlab.co.uk/open">Click here to learn more</a>
//           ___  _     _           _   _          _
//          / _ \| |__ (_) ___  ___| |_| |    __ _| |__
//         | | | | '_ \| |/ _ \/ __| __| |   / _` | '_ \
//         | |_| | |_) | |  __/ (__| |_| |__| (_| | |_) |
//          \___/|_.__// |\___|\___|\__|_____\__,_|_.__/
//                   |__/
//
//                   http://www.ObjectLab.co.uk
// ---------------------------------------------------------------------------
//
//QALab is released under the GNU General Public License.
//
//QALab: Collects QA Statistics from your build over time.
//2005+, ObjectLab Ltd
//
//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.
//
//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//General Public License for more details.
//
//You should have received a copy of the GNU General Public
//License along with this library; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
////////////////////////////////////////////////////////////////////////////////
package net.objectlab.qalab.parser;

import java.io.IOException;
import java.text.ParseException;
import java.util.Calendar;
import java.util.Date;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import net.objectlab.qalab.interfaces.QALabExporter;
import net.objectlab.qalab.util.QALabTags;
import net.objectlab.qalab.util.TaskLogger;

import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * This class abstract some methods used by all Stat Mergers being for
 * Checkstyle, pmd, findbugs or Simian. This class implements some SAX callbacks
 * for parsing the input statistics.
 * 
 * @author Benoit Xhenseval
 * @version $Revision: 187 $
 */
public abstract class BaseStatMerge extends DefaultHandler implements
        StatMerger {
    // ~ Instance fields
    // ------------------------------------------------------------------------

    /** total Statistics. * */
    private int totalStatistics = 0;

    /** count the number of files. * */
    private int fileCount = 0;

    /** statistics for given file. * */
    private int fileStatistics = 0;

    /** keep the current file name. * */
    private String currentFile = null;

    /** if true, no debug information is logged. * */
    private boolean quiet = true;

    /** given implementation for the logger. * */
    private TaskLogger taskLogger = null;

    /** The directory for source code. * */
    private String srcDir;

    /** exporter for the statistics. */
    private QALabExporter exporter = null;

    /** timestamp to use for Merge. */
    private Date mergerTimeStamp;

    /** timestamp as date only. */
    private boolean timeStampAsDateOnly = true;

    // ~ Methods
    // ------------------------------------------------------------------------

    /**
     * @param theFile
     *            the current file for which the stats are processed.
     */
    public final void setCurrentFile(final String theFile) {
        this.currentFile = theFile;
    }

    /**
     * @return true means that no debug info is logged.
     */
    public final boolean isQuiet() {
        return quiet;
    }

    /**
     * @param noLog
     *            true if no log required.
     */
    public final void setQuiet(final boolean noLog) {
        this.quiet = noLog;
    }

    /**
     * set the logger to use.
     * 
     * @param task
     *            the logger.
     */
    public final void setTaskLogger(final TaskLogger task) {
        this.taskLogger = task;
    }

    /**
     * @return logger to use if not quiet.
     */
    protected final TaskLogger getTaskLogger() {
        return taskLogger;
    }

    /**
     * this will be used to truncate machine specific path.
     * 
     * @param directory
     *            the source directory.
     */
    public final void setSrcDir(final String directory) {
        srcDir = directory.replace('\\', '/');
    }

    /**
     * @return the source directory
     */
    public final String getSrcDir() {
        return srcDir;
    }

    /**
     * Add new results to the compiled statistics. This could will generate a
     * new &lt;result&gt; element and may be a new &lt;file&gt; element.
     */
    protected final void addNewResults() {
        if (getFileStatistics() != 0) {
            exporter.addFileResult(getFileStatistics(), currentFile);
        } else {
            if (!isQuiet()) {
                taskLogger.log("No violation for " + currentFile);
            }
        }
    }

    /**
     * @return give the total number of violations.
     */
    public int getTotalStatistics() {
        return totalStatistics;
    }

    /**
     * @param increment
     *            add increment to total number of violations.
     */
    protected final void addTotalStatistics(final int increment) {
        totalStatistics += increment;
    }

    /**
     * Directly set the total statistics.
     * 
     * @param value
     */
    protected final void setTotalStatistics(final int value) {
        totalStatistics = value;
    }

    /**
     * @return number of file counted.
     */
    public final int getFileCount() {
        return fileCount;
    }

    /**
     * Increment file count by increment.
     * 
     * @param increment
     *            increase file count by this value.
     */
    protected final void incrementFileCount(final int increment) {
        fileCount += increment;
    }

    /**
     * @return number of violation for current file.
     */
    protected final int getFileStatistics() {
        return fileStatistics;
    }

    /**
     * add a increment of file violation and also add it to the total violation
     * count.
     * 
     * @param increment
     *            value to add to the number of fileStatistics.
     */
    protected final void addFileStatistics(final int increment) {
        fileStatistics += increment;
        addTotalStatistics(increment);
    }

    /**
     * reset the fileStatistics to 0, typically when handling a new file.
     */
    protected final void resetFileStatistics() {
        fileStatistics = 0;
    }

    /**
     * called by the ant task or the maven plug in.
     * 
     * @param source
     *            the Reader from the statistics file.
     * @param qalabExporter
     *            the QALabExporter to use for saving the stats.
     */
    public final void mergeStats(final InputSource source,
            final QALabExporter qalabExporter) {
        exporter = qalabExporter;

        try {
            final SAXParserFactory factory = SAXParserFactory.newInstance();
            final SAXParser saxParser = factory.newSAXParser();

            saxParser.getXMLReader().setEntityResolver(new EntityResolver() {
                public InputSource resolveEntity(String publicId,
                        String systemId) throws SAXException, IOException {
                    
                    return new InputSource();
                }
            });
            // As suggested by Nascif Abousalh-Neto, using an external DTD for
            // this may
            // cause some issues with Cobertura. See article for some features:
            // http://www-128.ibm.com/developerworks/xml/library/x-tipcfsx.html
            saxParser.getXMLReader().setFeature(
                    "http://xml.org/sax/features/external-general-entities",
                    false);
            saxParser.getXMLReader().setFeature(
                    "http://xml.org/sax/features/external-parameter-entities",
                    false);

            InputSource src = preProcessSource(source);
            
            saxParser.parse(src, this);
        } catch (ParserConfigurationException e) {
            getTaskLogger().log(e.toString());
        } catch (SAXException e) {
            getTaskLogger().log(e.toString());
        } catch (IOException e) {
            e.printStackTrace();
            getTaskLogger().log(e.toString());
        } finally {
            if (!isQuiet()) {
                getTaskLogger().log("Finished Parsing Source XML.");
            }
        }

        if (getTotalStatistics() != 0) {
            exporter.addSummary(getTotalStatistics(), getFileCount());
        }
    }

    protected InputSource preProcessSource(InputSource source) {
//        return new InputSource(source);
        return source;
    }

    /**
     * set the merger time stamp to use for all stats.
     * 
     * @param timestamp
     *            the timestamp "YYYY-MM-dd HH:mm:ss"
     * @param timeStampAsDate
     *            The timeStampAsDateOnly to set.
     */
    public final void setMergerTimeStamp(final String timestamp,
            final boolean timeStampAsDate) {
        this.timeStampAsDateOnly = timeStampAsDate;
        if (timestamp != null) {
            try {
                mergerTimeStamp = QALabTags.DEFAULT_DATETIME_FORMAT
                        .parse(timestamp);
            } catch (ParseException e) {
                throw new IllegalArgumentException(
                        "Wrong format for mergerTimeStamp. " + e.getMessage());
            }
        } else {
            mergerTimeStamp = new Date();
        }

        if (timeStampAsDate) {
            Calendar cal = Calendar.getInstance();
            cal.setTime(mergerTimeStamp);
            cal.set(Calendar.HOUR_OF_DAY, 0);
            cal.set(Calendar.MINUTE, 0);
            cal.set(Calendar.SECOND, 0);
            cal.set(Calendar.MILLISECOND, 0);
            mergerTimeStamp = cal.getTime();
        }
    }

    /**
     * This may have been changed to "now".
     * 
     * @return timestamp the time stamp in format YYYY-MM-dd HH:mm:ss
     */
    public final String getMergerTimeStamp() {
        if (mergerTimeStamp == null) {
            setMergerTimeStamp(null, timeStampAsDateOnly);
        }

        if (timeStampAsDateOnly) {
            return QALabTags.DEFAULT_DATE_FORMAT.format(mergerTimeStamp);
        } else {
            return QALabTags.DEFAULT_DATETIME_FORMAT.format(mergerTimeStamp);
        }
    }
}
/*
 *                   ObjectLab is sponsoring QALab
 * 
 * Based in London, we are world leaders in the design and development 
 * of bespoke applications for the securities financing markets.
 * 
 * <a href="http://www.objectlab.co.uk/open">Click here to learn more about us</a>
 *           ___  _     _           _   _          _
 *          / _ \| |__ (_) ___  ___| |_| |    __ _| |__
 *         | | | | '_ \| |/ _ \/ __| __| |   / _` | '_ \
 *         | |_| | |_) | |  __/ (__| |_| |__| (_| | |_) |
 *          \___/|_.__// |\___|\___|\__|_____\__,_|_.__/
 *                   |__/
 *
 *                     www.ObjectLab.co.uk
 */
