////////////////////////////////////////////////////////////////////////////////
//
//                  ObjectLab is sponsoring QALab
// 
// Based in London, we are world leaders in the design and development 
// of bespoke applications for the Securities Financing markets.
// 
// <a href="http://www.objectlab.co.uk/open">Click here to learn more</a>
//           ___  _     _           _   _          _
//          / _ \| |__ (_) ___  ___| |_| |    __ _| |__
//         | | | | '_ \| |/ _ \/ __| __| |   / _` | '_ \
//         | |_| | |_) | |  __/ (__| |_| |__| (_| | |_) |
//          \___/|_.__// |\___|\___|\__|_____\__,_|_.__/
//                   |__/
//
//                   http://www.ObjectLab.co.uk
// ---------------------------------------------------------------------------
//
//QALab is released under the GNU General Public License.
//
//QALab: Collects QA Statistics from your build over time.
//2005+, ObjectLab Ltd
//
//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.
//
//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//General Public License for more details.
//
//You should have received a copy of the GNU General Public
//License along with this library; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
////////////////////////////////////////////////////////////////////////////////
package net.objectlab.qalab.parser;

import net.objectlab.qalab.util.QALabTags;
import net.objectlab.qalab.util.TaskLogger;
import net.objectlab.qalab.util.Util;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import java.io.IOException;
import java.io.Writer;

import java.text.ParseException;

import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Stack;

/**
 * A handler class which uses SAX method to parse the XML file from QALab and
 * extract the files that have seen a change in statistics within the time
 * window.
 * 
 * @author Paramjit Rehinsi
 * @version $Revision: 199 $
 * @todo add a StartDate for the window
 * @todo add an EndDate for the Window
 * @todo simplify the constructor to remove File dependencies.
 */
public class BuildStatMoverHandler extends DefaultHandler {
    private static final int ONE_DAY = 24;

    /**
     * default 48h to deduct is weekendAdjustment.
     */
    private static final int WEEKEND_SKIP = 48;

    // ~ Instance fields
    // ------------------------------------------------------------------------

    /**
     * The types to be handled checkstyle|findbugs|pmd|simian.
     */
    private List listOfTypes = new LinkedList();

    /**
     * The logger to use for exceptions and debug information.
     */
    private TaskLogger tasklogger;

    /**
     * The QALab input source.
     */
    private InputSource qalabSource;

    /**
     * the XML writer with the "movers".
     */
    private Writer outputWriter;

    /**
     * A map containing a Stack of items per type.
     */
    private Map mapOfTypes = new HashMap();

    /**
     * list of file results.
     */
    private List filebeans = new LinkedList();

    /**
     * the current file result.
     */
    private FileStats fileResult;

    /**
     * the current file name.
     */
    private String filename;

    /**
     * the current path to file.
     */
    private String path;

    /**
     * cutoff earliest date to consider in the stats.
     */
    private Date startDateCutoff;

    /**
     * cutoff latest date to consider in the stats.
     */
    private Date endDateCutoff;

    /**
     * StartTimeWindow.
     */
    private Date startTimeWindow = null;

    /**
     * End Time Window.
     */
    private Date endTimeWindow = null;

    /**
     * start time offset (from now or endTimeWindow).
     */
    private String offsetHours;

    /**
     * weekendAdjustment, true and an extra 48 will be subtracted to any
     * datecutoff that fell on a Saturday or Sunday.
     */
    private boolean wkendadj = false;

    /**
     * quiet no debug if true (default).
     */
    private boolean quiet = true;
    
    /** offset in hours to determine if the stat has fallen to 0 over the last n hours
     * from the endDateCutoff.
     */
    private int hoursOffsetForLastRun = ONE_DAY;

    // ~ Constructors
    // ------------------------------------------------------------------------

    /**
     * Constructor used by ant Task/Maven.
     * 
     * @param qaLabSource
     *            the original QALab file.
     * @param types
     *            the types to extract.
     * @param outputFile
     *            the output file.
     * @param weekendAdjustment
     *            true if it should adjust for weekends.
     * @param log
     *            the logger to use.
     * @param noDebug
     *            if true, quiet output (no debug info).
     */
    public BuildStatMoverHandler(final InputSource qaLabSource,
            final String types, final Writer outputFile,
            final boolean weekendAdjustment, final TaskLogger log,
            final boolean noDebug) {
        setLogger(log);
        quiet = noDebug;
        wkendadj = weekendAdjustment;
        listOfTypes = Util.listify(types, ",");
        processTypes(listOfTypes);
        this.outputWriter = outputFile;

        qalabSource = qaLabSource;
    }

    // ~ Methods
    // ------------------------------------------------------------------------

    /**
     * Create an instance of the SAX Parser and process the qalab.xml file.
     * 
     * @throws ParserConfigurationException
     *             cannot create an instance of parser.
     * @throws SAXException
     *             any SAX issue.
     * @throws java.io.IOException
     *             any IO/access to file issue.
     */
    public final void process() throws ParserConfigurationException,
            SAXException, IOException {
        calculateCutoffDateTime();

        SAXParserFactory factory = SAXParserFactory.newInstance();
        SAXParser saxParser = factory.newSAXParser();

        saxParser.parse(getQalabSource(), this);
    }

    /**
     * Process the types, assign a new stack to each type.
     * 
     * @param listoftypes
     *            list of types to support.
     */
    private void processTypes(final List listoftypes) {
        Iterator i = listoftypes.iterator();

        while (i.hasNext()) {
            mapOfTypes.put(i.next(), new Stack());
        }
    }

    /**
     * Gets the filename from the id attribute.
     * 
     * @param attrs
     *            XML attributes.
     */
    public final void getFileName(final Attributes attrs) {
        if (attrs != null) {
            for (int i = 0; i < attrs.getLength(); i++) {
                String name = attrs.getQName(i); // Attr name
                String value = attrs.getValue(i);

                if (QALabTags.ID_ATTRIBUTE.equals(name)) {
                    filename = value;

                    if (!quiet) {
                        tasklogger.log("File:" + filename);
                    }
                } else if (QALabTags.PATH_ATTRIBUTE.equals(name)) {
                    path = value;
                }
            }
        }
    }

    /**
     * Process the attributes of each result and return a SingleStat object
     * built from the attributes date, type and statvalue.
     * 
     * @param attrs
     *            XML attributes
     * @return new SingleStat object built from the attributes.
     */
    private SingleStat buildResultBean(final Attributes attrs) {
        SingleStat singleStat = new SingleStat();

        if (attrs != null) {
            for (int i = 0; i < attrs.getLength(); i++) {
                String name = attrs.getQName(i); // Attr name
                String value = attrs.getValue(i);

                if (QALabTags.TYPE_ATTRIBUTE.equals(name)) {
                    singleStat.setType(value);
                } else if (QALabTags.STATVALUE_ATTRIBUTE.equals(name)) {
                    singleStat.setStatValue(Integer.parseInt(value));
                } else if (QALabTags.DATE_ATTRIBUTE.equals(name)) {
                    try {
                        if (value != null
                                && value.length() > QALabTags.DATE_ONLY_SIZE) {
                            singleStat
                                    .setDate(QALabTags.DEFAULT_DATETIME_FORMAT
                                            .parse(value));
                        } else if (value != null
                                && value.length() == QALabTags.DATE_ONLY_SIZE) {
                            singleStat.setDate(QALabTags.DEFAULT_DATE_FORMAT
                                    .parse(value));
                        }
                    } catch (ParseException e) {
                        tasklogger.log(e.toString());
                    }
                }
            }
        }

        return singleStat;
    }

    // ===========================================================
    // SAX DocumentHandler methods
    // ===========================================================

    /**
     * At the start of a new element, capture the filename and if the element is
     * a result one, create a SingleStat to store for the given type.
     * 
     * @param ignoreNamespaceURI
     *            ignore (present for interface implementation).
     * @param localname
     *            name of the current element.
     * @param qualifiedname
     *            element name.
     * @param attrs
     *            the XML attribute of the current element.
     * @throws SAXException
     *             any SAX issue
     */
    public final void startElement(final String ignoreNamespaceURI, // NOPMD
            final String localname, final String qualifiedname,
            final Attributes attrs) throws SAXException {
        String local = localname;
        if ("".equals(local)) {
            local = qualifiedname;
        }

        // if (!quiet) {
        // tasklogger.log("Start :" + qualifiedname);
        // }

        if (QALabTags.FILE_TAG.equals(local)) {
            getFileName(attrs);
        }

        if (QALabTags.RESULT_TAG.equals(local)) {
            SingleStat resultbean = buildResultBean(attrs);
            Stack s = (Stack) mapOfTypes.get(resultbean.getType());

            // only push items that are within the time window.
            if (s != null) {
                s.push(resultbean);
            }
        }
    }

    /**
     * At the end of an element, check if it is a file one and add the results
     * found.
     * 
     * @param ignoreNamespaceURI
     *            ignore (present for interface implementation).
     * @param ignoreSimplename
     *            ignore (present for interface implementation).
     * @param qualifiedname
     *            the name of the element.
     * @throws SAXException
     *             any SAX issue
     */
    public final void endElement(final String ignoreNamespaceURI, // NOPMD
            final String ignoreSimplename, final String qualifiedname) // NOPMD
            throws SAXException {
        if (QALabTags.FILE_TAG.equals(qualifiedname)) {
            fileResult = new FileStats(mapOfTypes, filename, path,
                    startDateCutoff);
            fileResult.setEndDateCutoff(endDateCutoff);
            fileResult.setOffsetForLastRun(hoursOffsetForLastRun);
            filebeans.add(fileResult);
            mapOfTypes = new HashMap();
            processTypes(listOfTypes);
        }
    }

    /**
     * Return the input source to the base stats.
     * 
     * @return InputSource typically the source to qalab.xml
     */
    public final InputSource getQalabSource() {
        return qalabSource;
    }

    /**
     * At the end of parsing the qalab.xml document, generate the up/down
     * report.
     * 
     * @throws SAXException
     *             any SAX issue.
     */
    public final void endDocument() throws SAXException {
        try {
            // now write the results
            outputWriter.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!DOCTYPE moversreport [");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ELEMENT moversreport (types,datethreshold,"
                    + "enddatethreshold,daterun,up?,down?)>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ELEMENT types (#PCDATA)>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ELEMENT datethreshold (#PCDATA)>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ELEMENT enddatethreshold (#PCDATA)>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ELEMENT daterun (#PCDATA)>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ELEMENT up (file+)>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ELEMENT down (file+)>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ELEMENT file (diff+)>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ATTLIST file name CDATA #REQUIRED>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ATTLIST file path CDATA #REQUIRED>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ELEMENT diff EMPTY>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ATTLIST diff previousrun CDATA #REQUIRED>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ATTLIST diff currentrun CDATA #REQUIRED>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ATTLIST diff type CDATA #REQUIRED>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ATTLIST diff previouserrors "
                    + "CDATA #REQUIRED>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ATTLIST diff currenterrors CDATA "
                    + "#REQUIRED>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<!ATTLIST diff diff CDATA #REQUIRED>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("]>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("<moversreport>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write(" <types>");
            outputWriter.write(Util.listToCSVString(listOfTypes));
            outputWriter.write("</types>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write(" <datethreshold>");
            outputWriter.write(QALabTags.DEFAULT_DATETIME_FORMAT
                    .format(startDateCutoff));
            outputWriter.write("</datethreshold>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write(" <enddatethreshold>");
            outputWriter.write(QALabTags.DEFAULT_DATETIME_FORMAT
                    .format(endDateCutoff));
            outputWriter.write("</enddatethreshold>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write(" <daterun>");
            outputWriter.write(QALabTags.DEFAULT_DATETIME_FORMAT
                    .format(new Date()));
            outputWriter.write("</daterun>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write(" <up>");
            outputWriter.write(QALabTags.LINE_END);

            for (Iterator it = filebeans.iterator(); it.hasNext();) {
                FileStats fb = (FileStats) it.next();

                fb.constructXML();
                outputWriter.write(fb.getUpXmlResult());
            }

            outputWriter.write(" </up>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write(" <down>");
            outputWriter.write(QALabTags.LINE_END);

            for (Iterator it = filebeans.iterator(); it.hasNext();) {
                FileStats fb = (FileStats) it.next();

                outputWriter.write(fb.getDownXmlResult());
            }

            outputWriter.write(" </down>");
            outputWriter.write(QALabTags.LINE_END);
            outputWriter.write("</moversreport>");

            outputWriter.flush();
        } catch (IOException e) {
            tasklogger.log(e.toString());
        } finally {
            if (outputWriter != null) {
                try {
                    outputWriter.close();
                } catch (IOException e) {
                    tasklogger.log(e.toString());
                }
            }
        }
    }

    /**
     * calculates the datecutoff by adding the number of offset hours.
     * 
     * @param offsetInHours
     *            number of hours from now.
     */
    public final void setDatecutoff(final int offsetInHours) {
        Calendar cal = Calendar.getInstance();

        cal.add(Calendar.HOUR, offsetInHours);
        startDateCutoff = cal.getTime();
    }

    /**
     * Calculate the cutoff date, ie the earliest date for the stat to be taken
     * into account. extra 48h if original cutoff date falls on a Saturday or
     * Sunday.
     */
    public final void calculateCutoffDateTime() {
        if (startTimeWindow != null) {
            startDateCutoff = startTimeWindow;
        } else {
            // always do the cut off
            int offset = -Integer.parseInt(offsetHours);

            setDatecutoff(offset);
        }

        if (wkendadj) {
            // check if weekend:
            Calendar cal = Calendar.getInstance();

            cal.setTime(startDateCutoff);

            final int dayOfWeek = cal.get(Calendar.DAY_OF_WEEK);

            if ((dayOfWeek == Calendar.SUNDAY)
                    || (dayOfWeek == Calendar.SATURDAY)) {
                cal.add(Calendar.HOUR, -WEEKEND_SKIP);
                startDateCutoff = cal.getTime();
            }
        }

        if (!quiet) {
            tasklogger.log("Cutoff = " + startDateCutoff);
        }

        if (endTimeWindow != null) {
            endDateCutoff = endTimeWindow;
        } else {
            endDateCutoff = new Date();
        }
    }

    /**
     * @param log
     *            the logger to use.
     */
    private void setLogger(final TaskLogger log) {
        this.tasklogger = log;
    }

    /**
     * @param endTime
     *            The endTimeWindow to set.
     */
    public final void setEndTimeWindow(final Date endTime) {
        this.endTimeWindow = new Date(endTime.getTime());
    }

    /**
     * @param startTime
     *            The startTimeWindow to set.
     */
    public final void setStartTimeWindow(final Date startTime) {
        this.startTimeWindow = new Date(startTime.getTime());
    }

    /**
     * @param offset
     *            The offsetHours to set.
     */
    public final void setOffsetHours(final String offset) {
        this.offsetHours = offset;
    }

    /**
     * @return Returns the offsetHours.
     */
    public final String getOffsetHours() {
        return offsetHours;
    }

    public int getHoursOffsetForLastRun() {
        return hoursOffsetForLastRun;
    }

    public void setHoursOffsetForLastRun(int hoursOffsetForLastRun) {
        this.hoursOffsetForLastRun = hoursOffsetForLastRun;
    }
}
/*
 *                   ObjectLab is sponsoring QALab
 * 
 * Based in London, we are world leaders in the design and development 
 * of bespoke applications for the securities financing markets.
 * 
 * <a href="http://www.objectlab.co.uk/open">Click here to learn more about us</a>
 *           ___  _     _           _   _          _
 *          / _ \| |__ (_) ___  ___| |_| |    __ _| |__
 *         | | | | '_ \| |/ _ \/ __| __| |   / _` | '_ \
 *         | |_| | |_) | |  __/ (__| |_| |__| (_| | |_) |
 *          \___/|_.__// |\___|\___|\__|_____\__,_|_.__/
 *                   |__/
 *
 *                     www.ObjectLab.co.uk
 */
