////////////////////////////////////////////////////////////////////////////////
//
//                  ObjectLab is sponsoring QALab
// 
// Based in London, we are world leaders in the design and development 
// of bespoke applications for the Securities Financing markets.
// 
// <a href="http://www.objectlab.co.uk/open">Click here to learn more</a>
//           ___  _     _           _   _          _
//          / _ \| |__ (_) ___  ___| |_| |    __ _| |__
//         | | | | '_ \| |/ _ \/ __| __| |   / _` | '_ \
//         | |_| | |_) | |  __/ (__| |_| |__| (_| | |_) |
//          \___/|_.__// |\___|\___|\__|_____\__,_|_.__/
//                   |__/
//
//                   http://www.ObjectLab.co.uk
// ---------------------------------------------------------------------------
//
//QALab is released under the GNU General Public License.
//
//QALab: Collects QA Statistics from your build over time.
//2005+, ObjectLab Ltd
//
//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.
//
//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//General Public License for more details.
//
//You should have received a copy of the GNU General Public
//License along with this library; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
////////////////////////////////////////////////////////////////////////////////
package net.objectlab.qalab.parser;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.io.StringReader;
import java.math.BigDecimal;

import net.objectlab.qalab.util.FilenameUtil;
import net.objectlab.qalab.util.Util;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * This class is able to read the XML output from Cobertura (LINE INFORMATION)
 * and integrate the violation results into our "qalab" xml format.
 * 
 * @author Benoit Xhenseval
 * @version $Revision$
 */
public class CoberturaLineStatMerge extends BaseStatMerge {
    private static final String COBERTURA_LINE = "cobertura-line";

    private int totalLineRate = 0;

    private int totalBranchRate = 0;

    private boolean skipCurrentElement = false;

    // ~ Methods
    // ------------------------------------------------------------------------

    /**
     * @return cobertura-line
     */
    public String getType() {
        return COBERTURA_LINE;
    }

    protected final String getSummaryTag() {
        return "coverage";
    }

    private Object getFileTag() {
        return "class";
    }

    /**
     * 
     * @return "filename" the attribute tag for file.
     */
    private String getFileNameAttribute() {
        return "filename";
    }

    /**
     * At the start of a new element, capture the filename and if the element is
     * a result one, create a SingleStat to store for the given type.
     * 
     * @param ignoreNamespaceURI
     *            ignore (present for interface implementation).
     * @param localname
     *            name of the current element.
     * @param qualifiedname
     *            element name.
     * @param attrs
     *            the XML attribute of the current element.
     * @throws SAXException
     *             any SAX issue
     */
    public final void startElement(final String ignoreNamespaceURI, // NOPMD
            final String localname, final String qualifiedname,
            final Attributes attrs) throws SAXException {
        String local = localname;

        if ("".equals(local)) {
            local = qualifiedname;
        }

        // if (!isQuiet()) {
        // getTaskLogger().log("Start :" + local);
        // }

        if (getSummaryTag().equals(local)) {
            final String lineRate = Util.getAttributeValue(attrs, "line-rate",
                    isQuiet(), getTaskLogger());
            final String branchRate = Util.getAttributeValue(attrs,
                    "branch-rate", isQuiet(), getTaskLogger());
            totalLineRate = new BigDecimal(lineRate).movePointRight(2)
                    .intValue();
            totalBranchRate = new BigDecimal(branchRate).movePointRight(2)
                    .intValue();
        } else if (getFileTag().equals(local)) {
            // check if it is an inner class... and if so, leave it!!
            String className = Util.getAttributeValue(attrs, "name", isQuiet(),
                    getTaskLogger());

            if (className != null && className.indexOf('$') > 0) {
                skipCurrentElement = true;
                return;
            }

            // this is the start of a new file.
            resetFileStatistics();
            incrementFileCount(1);
            setFileName(attrs);

            final String lineRate = Util.getAttributeValue(attrs, "line-rate",
                    isQuiet(), getTaskLogger());
            final String branchRate = Util.getAttributeValue(attrs,
                    "branch-rate", isQuiet(), getTaskLogger());

            if (!isQuiet()) {
                getTaskLogger().log(
                        "lineRate=" + lineRate + " branchRate=" + branchRate);
            }

            if (getType().equals(COBERTURA_LINE)) {
                int lineAsPercent = new BigDecimal(lineRate).movePointRight(2)
                        .intValue();
                addFileStatistics(lineAsPercent);
            } else {
                // branches
                int lineAsPercent = new BigDecimal(branchRate)
                        .movePointRight(2).intValue();
                addFileStatistics(lineAsPercent);
            }
        }
    }

    /**
     * At the end of an element, check if it is a file one and add the results
     * found.
     * 
     * @param ignoreNamespaceURI
     *            ignore (present for interface implementation).
     * @param ignoreSimplename
     *            ignore (present for interface implementation).
     * @param qualifiedname
     *            the name of the element.
     * @throws SAXException
     *             any SAX issue
     */
    public final void endElement(final String ignoreNamespaceURI, // NOPMD
            final String ignoreSimplename, final String qualifiedname) // NOPMD
            throws SAXException {
        // if (!isQuiet()) {
        // getTaskLogger().log("End :" + qualifiedname);
        // }

        if (getFileTag().equals(qualifiedname)) {
            if (!skipCurrentElement) {
                addNewResults();
            } else {
                skipCurrentElement = false;
            }
        }
    }

    /**
     * set the file name from the attributes from Findbugs XML.
     * 
     * @param att
     *            xml attributes
     */
    protected void setFileName(final Attributes att) {
        String fileName = Util.getAttributeValue(att, getFileNameAttribute(),
                isQuiet(), getTaskLogger());

        // if (!isQuiet()) {
        // getTaskLogger().log("File:" + fileName);
        // }

        fileName = FilenameUtil.trimFilename(fileName, getSrcDir());

        if (!isQuiet()) {
            getTaskLogger().log(
                    "setCurrentFileName FILE [" + fileName + "] srcDir ["
                            + getSrcDir() + "]");
        }

        setCurrentFile(fileName);
    }

    /**
     * For Cobertura the total statistics is the average of coverage, ie sum of
     * all percentages divided by the TOTAL number of files (ie we do not report
     * ZERO % coverage but they should count).
     */
    public int getTotalStatistics() {
        if (getType().equals(COBERTURA_LINE)) {
            return totalLineRate;
        } else {
            return totalBranchRate;
        }
    }

    /**
     * This is required to get rid of the SYSTEM DTD otherwise one cannot parse
     * coverage.xml behind most firewalls or offline.
     */
    protected InputSource preProcessSource(InputSource source) {
        StringBuffer buf = new StringBuffer();
        InputStream reader = source.getByteStream();
        int i = -1;
        // the InputSource could be either an InputStream or a Reader
        // we need to cope with both...
        if (reader != null) {
            do {
                try {
                    i = reader.read();
                } catch (IOException e) {
                    getTaskLogger().log(e.toString());
                }
                buf.append((char) i);
            } while (i >= 0);
        } else {
            Reader rd = source.getCharacterStream();
            do {
                try {
                    i = rd.read();
                } catch (IOException e) {
                    getTaskLogger().log(e.toString());
                }
                buf.append((char) i);
            } while (i >= 0);
        }
        int idx = buf.indexOf("<coverage");
        if (idx >= 0) {
            buf = buf.delete(0, idx);
        }

        return new InputSource(new StringReader(buf.toString()));
    }
}
/*
 *                   ObjectLab is sponsoring QALab
 * 
 * Based in London, we are world leaders in the design and development 
 * of bespoke applications for the securities financing markets.
 * 
 * <a href="http://www.objectlab.co.uk/open">Click here to learn more about us</a>
 *           ___  _     _           _   _          _
 *          / _ \| |__ (_) ___  ___| |_| |    __ _| |__
 *         | | | | '_ \| |/ _ \/ __| __| |   / _` | '_ \
 *         | |_| | |_) | |  __/ (__| |_| |__| (_| | |_) |
 *          \___/|_.__// |\___|\___|\__|_____\__,_|_.__/
 *                   |__/
 *
 *                     www.ObjectLab.co.uk
 */
